from aquests.athreads import trigger
from aquests import lifetime
import sys, time
from rs4 import asyncore
import gc
import select
import os
import bisect
import socket
import time

if os.name == "nt":
	from errno import WSAENOTSOCK

_shutdown_phase = 0
_shutdown_timeout = 30 # seconds per phase
_exit_code = 0
_last_maintern = 0
_maintern_interval = 3.0
_killed_zombies = 0
_select_errors = 0

def status ():
	fds = []
	for fdno, channel in list(asyncore.socket_map.items ()):
		d = {}
		d ["name"] = "%s.%s" % (channel.__module__, channel.__class__.__name__)
		d ["fdno"] = fdno

		status = "NOTCON"
		if channel.accepting and channel.addr: status = "LISTEN"
		elif channel.connected: status = "CONNECTED"
		d ["status"] = status
		addr = ""
		if channel.addr is not None:
			try: addr = "%s:%d" % channel.addr
			except TypeError: addr = "%s" % repr (channel.addr)
		if addr:
			d ["address"] = addr

		if hasattr (channel, "channel_number"):
			d ["channel_number"] = channel.channel_number
		if hasattr (channel, "request_counter"):
			d ["request_counter"] = channel.request_counter
		if hasattr (channel, "event_time"):
			d ["last_event_time"] = time.asctime (time.localtime (channel.event_time))
		if hasattr (channel, "zombie_timeout"):
			d ["zombie_timeout"] = channel.zombie_timeout
		if hasattr (channel, "get_history"):
			d ["history"] = channel.get_history ()

		fds.append (d)

	return {
		"killed_zombies": _killed_zombies,
		"select_errors": _select_errors,
		"selecting_sockets": len (asyncore.socket_map),
		"in_the_map": fds
	}

def maintern_zombie_channel (now):
	global _killed_zombies

	for channel in list(asyncore.socket_map.values()):
		if hasattr (channel, "handle_timeout"):
			try:
				iszombie = (now - channel.event_time) > channel.zombie_timeout + 3
			except AttributeError:
				continue
			if iszombie:
				_killed_zombies += 1
				try:
					channel.handle_timeout ()
				except:
					channel.handle_error ()

maintern = lifetime.Maintern ()
tick_timer = lifetime.TickTimer ()
def init (kill_zombie_interval = 60.0, logger = None):
	global maintern
	lifetime.EXHAUST_DNS = False
	lifetime._logger = logger
	maintern.sched (kill_zombie_interval, lifetime.maintern_zombie_channel)
	maintern.sched (60.0, maintern_gc)
	if "---gc" in sys.argv:
		gc.set_debug (gc.DEBUG_SAVEALL)

def maintern_gc (now):
	gc.collect ()
	if "---gc" in sys.argv:
		print ("# Garbages collected ===========================")
		for item in gc.garbage:
			print ("  -", item)

def manual_gc (interval = 60.0):
	global maintern
	maintern.sched (interval, maintern_gc)
	gc.disable ()

def enable_memory_track (interval = 10.0):
	global maintern
	maintern.sched (interval, lifetime.summary_track)

def shutdown (exit_code, shutdown_timeout = 30.0):
	global _shutdown_phase
	global _shutdown_timeout
	global _exit_code

	if _shutdown_phase:
		# aready entered
		return

	if _shutdown_phase == 0:
		_exit_code = exit_code
		_shutdown_phase = 1

	_shutdown_timeout = shutdown_timeout

	trigger.wakeselect ()

def loop (timeout = 30.0):
	global _shutdown_phase
	global _shutdown_timeout
	global _exit_code
	global maintern

	if not maintern.q:
		init ()

	_shutdown_phase = 0
	_shutdown_timeout = 30
	_exit_code = 0

	try:
		lifetime_loop (timeout)
	except KeyboardInterrupt:
		_shutdown_timeout = 1
		graceful_shutdown_loop()
	else:
		graceful_shutdown_loop()

def lifetime_loop (timeout = 30.0, map = None):
	global _last_maintern
	global _maintern_interval

	map = map or asyncore.socket_map
	while map and _shutdown_phase == 0:
		lifetime.poll_fun_wrap (timeout, map)
		tick_timer.tick ()
		now = time.time()
		if (now - _last_maintern) > _maintern_interval:
			maintern and maintern (now)
			_last_maintern = time.time ()

def graceful_shutdown_loop ():
	global _shutdown_phase
	timestamp = time.time()
	timeout = 1.0
	map = asyncore.socket_map
	while map and _shutdown_phase < 4:
		time_in_this_phase = time.time() - timestamp
		veto = 0
		for fd,obj in list(map.items()):
			try:
				fn = getattr (obj,'clean_shutdown_control')
			except AttributeError:
				pass
			else:
				try:
					veto = veto or fn (_shutdown_phase, time_in_this_phase)
				except:
					obj.handle_error()

		if veto and time_in_this_phase < _shutdown_timeout:
			lifetime.poll_fun_wrap (timeout, map)
		else:
			_shutdown_phase += 1
			timestamp = time.time()



import time
from aquests.athreads import socket_map
from aquests.athreads import trigger
from rs4.cbutil import tuple_cb
from aquests.client.asynconnect import AsynSSLConnect, AsynConnect
from aquests.dbapi.dbconnect import DBConnect
import threading
from aquests.protocols.http import request as http_request
from aquests.protocols.http import request_handler as http_request_handler
from aquests.protocols.http2 import request_handler as http2_request_handler
from aquests.protocols.grpc.request import GRPCRequest
from aquests.protocols.http import response as http_response
from aquests.protocols.ws import request_handler as ws_request_handler
from aquests.protocols.ws import request as ws_request
from . import rcache
from skitai import lifetime
from aquests import asyncore
import sys
import inspect
from skitai import exceptions
from skitai import REQFAIL, UNSENT, TIMEOUT, NETERR, NORMAL
from ...corequest import corequest, response
import sqlite3

try:
    import psycopg2
except ImportError:
    class PGIntegrityError (Exception):
        pass
else:
    PGIntegrityError = psycopg2.IntegrityError

DEFAULT_TIMEOUT = 10
WAIT_POLL = False

class OperationError (Exception):
    pass


class Result (response, rcache.Result):
    def __init__ (self, id, status, response, ident = None):
        rcache.Result.__init__ (self, status, ident)
        self.node = id
        self.__response = response

    def __getattr__ (self, attr):
        return getattr (self.__response, attr)

    def reraise (self):
        if self.status_code >= 300:
            try:
                self.__response.expt
            except AttributeError:
                # redircting to HTTPError
                raise exceptions.HTTPError ("%d %s" % (self.status_code, self.reason))
            else:
                self.__response.raise_for_status ()
        return self

    def close (self):
        self.__response = None

    def cache (self, timeout = 60, cache_if = (200,)):
        if not timeout:
            return
        if self.status != NORMAL or self.status_code not in cache_if:
            return
        rcache.Result.cache (self, timeout)
        return self

    def fetch (self, cache = None, cache_if = (200,), one = False):
        self.reraise ()
        self.cache (cache, cache_if)
        if one:
            if len (self.data) == 0:
                raise exceptions.HTTPError ("410 Partial Not Found")
            if len (self.data) != 1:
                raise exceptions.HTTPError ("409 Conflict")
            if isinstance (self.data, dict):
                return self.data.popitem () [1]
            return self.data [0]
        return self.data

    def one (self, cache = None, cache_if = (200,)):
        try:
            return self.fetch (cache, cache_if, True)
        except (PGIntegrityError, sqlite3.IntegrityError):
            # primary or unique index error
            raise exceptions.HTTPError ("409 Conflict")

    def commit (self):
        self.reraise ()


class Results (response, rcache.Result):
    def __init__ (self, results, ident = None):
        self.results = results
        self.status_code = [rs.status_code for rs in results]
        rcache.Result.__init__ (self, [rs.status for rs in self.results], ident)

    def __iter__ (self):
        return self.results.__iter__ ()

    @property
    def data (self):
        return [r.data for r in self.results]

    @property
    def text (self):
        return [r.text for r in self.results]

    def reraise (self):
        [r.reraise () for r in self.results]

    def cache (self, timeout = 60, cache_if = (200,)):
        if [_f for _f in [rs.status != NORMAL or rs.status_code not in cache_if for rs in self.results] if _f]:
            return
        rcache.Result.cache (self, timeout)
        return self

    def fetch (self, cache = None, cache_if = (200,)):
        self.cache (cache, cache_if)
        return [r.fetch () for r in self.results]

    def one (self, cache = None, cache_if = (200,)):
        self.cache (cache, cache_if)
        return [r.one () for r in self.results]



class Dispatcher:
    def __init__ (self, cv, id, ident = None, filterfunc = None, cachefs = None, callback = None):
        self._cv = cv
        self.id = id
        self.ident = ident
        self.filterfunc = filterfunc
        self.cachefs = cachefs
        self.callback = callback
        self.creation_time = time.time ()
        self.status = UNSENT
        self.result = None
        self.handler = None

    def get_id (self):
        return self.id

    def get_status (self):
        with self._cv:
            return self.status

    def request_failed (self):
        self.status = REQFAIL
        tuple_cb (self, self.callback)

    def set_status (self, code, result = None):
        with self._cv:
            self.status = code
            if result:
                self.result = result
        return code

    def get_result (self):
        if not self.result:
            if self.get_status () == REQFAIL:
                self.result = Result (self.id, REQFAIL, http_response.FailedResponse (731, "Request Failed"), self.ident)
            else:
                self.result = Result (self.id, TIMEOUT, http_response.FailedResponse (730, "Timeout"), self.ident)
        return self.result

    def do_filter (self):
        if self.filterfunc:
            self.filterfunc (self.result)

    def handle_cache (self, response):
        self.set_status (NORMAL, Result (self.id, status, response, self.ident))

    def handle_result (self, handler):
        if self.get_status () == TIMEOUT:
            # timeout, ignore
            return

        response = handler.response
        # DON'T do_filter here, it blocks select loop
        if response.code >= 700:
            if response.code == 702:
                status = TIMEOUT
            else:
                status = NETERR
        else:
            status = NORMAL

        result = Result (self.id, status, response, self.ident)

        cakey = response.request.get_cache_key ()
        if self.cachefs and cakey and response.max_age:
            self.cachefs.save (
                cakey,
                response.get_header ("content-type"), response.content,
                response.max_age, 0
            )

        handler.callback = None
        handler.response = None
        self.set_status (status, result)
        tuple_cb (self, self.callback)


class Task (corequest):
    DEFAULT_CACHE_TIMEOUT = 42
    proto_map = dict (
       rpc = http_request.XMLRPCRequest,
       xmlrpc = http_request.XMLRPCRequest,
       jsonrpc = http_request.JSONRPCRequest,
       grpc = GRPCRequest
    )
    def __init__ (self,
        cluster,
        uri,
        params = None,
        reqtype = "get",
        headers = None,
        auth = None,
        meta = None,
        use_cache = False,
        mapreduce = True,
        filter = None,
        callback = None,
        cache = None,
        timeout = 10,
        origin = None,
        cachefs = None,
        logger = None
        ):

        self._uri = uri
        self._params = params
        self._headers = headers
        self._reqtype = reqtype
        self._auth = auth
        self.set_defaults (cluster, meta, use_cache, mapreduce, filter, callback, cache, timeout, origin, logger, cachefs)

        if not self._reqtype.lower ().endswith ("rpc"):
            self._build_request ("", self._params)

    @classmethod
    def add_proto (cls, name, class_):
        cls.proto_map [name] = class_

    def set_defaults (self, cluster, meta, use_cache, mapreduce, filter, callback, cache, timeout, origin, logger, cachefs = None):
        self._cluster = cluster
        self._meta = meta or {}
        self._use_cache = use_cache
        self._mapreduce = mapreduce
        self._filter = filter
        self._callback = callback
        self._cache_timeout = cache
        self._timeout = timeout
        self._origin = origin
        self._cachefs = cachefs
        self._logger = logger
        self._requests = {}
        self._results = []
        self._canceled = False
        self._init_time = time.time ()
        self._cv = None
        self._retry = 0
        self._numnodes = 0
        self._cached_result = None
        self._cached_request_args = None
        self._request = None
        self._ccv = None

        if self._cluster:
            nodes = self._cluster.get_nodes ()
            self._numnodes = len (nodes)
            if self._mapreduce:
                self._nodes = nodes
            else: # anyone of nodes
                self._nodes = [None]

    def __del__ (self):
        self._cv = None
        self._results = []

    def _get_ident (self):
        cluster_name = self._cluster.get_name ()
        if cluster_name == "__socketpool__":
            _id = "%s/%s" % (self._uri, self._reqtype)
        else:
            _id = "%s/%s/%s" % (cluster_name, self._uri, self._reqtype)
        _id += "/%s/%s" % self._cached_request_args
        _id += "%s" % (
            self._mapreduce and "/M" or ""
            )
        return _id

    def _add_header (self, n, v):
        if self._headers is None:
            self._headers = {}
        self._headers [n] = v

    def _handle_request (self, request, rs, asyncon, handler):
        if self._cachefs:
            # IMP: mannual address setting
            request.set_address (asyncon.address)
            cakey = request.get_cache_key ()
            if cakey:
                cachable = self._cachefs.is_cachable (
                    request.get_header ("cache-control"),
                    request.get_header ("cookie") is not None,
                    request.get_header ("authorization") is not None,
                    request.get_header ("pragma")
                )

                if cachable:
                    hit, compressed, max_age, content_type, content = self._cachefs.get (cakey, undecompressible = 0)
                    if hit:
                        header = "HTTP/1.1 200 OK\r\nContent-Type: %s\r\nX-Skitaid-Cache-Lookup: %s" % (
                            content_type, hit == 1 and "MEM_HIT" or "HIT"
                        )
                        response = http_response.Response (request, header)
                        response.collect_incoming_data (content)
                        response.done ()
                        asyncon.set_active (False)
                        rs.handle_cache (response)
                        return 0

        r = handler (asyncon, request, rs.handle_result)
        if asyncon.get_proto () and asyncon.isconnected ():
            asyncon.handler.handle_request (r)
        else:
            r.handle_request ()

        return 1

    def _build_request (self, method, params):
        self._cached_request_args = (method, params) # backup for retry
        if self._use_cache and rcache.the_rcache:
            self._cached_result = rcache.the_rcache.get (self._get_ident (), self._use_cache)
            if self._cached_result is not None:
                self._cached_result.meta = self._meta
                self._callback and tuple_cb (self._cached_result, self._callback)
                return
            else:
                self._use_cache = False

        requests = 0
        while self._avails ():
            if self._cluster.get_name () != "__socketpool__":
                asyncon = self._get_connection (None)
            else:
                asyncon = self._get_connection (self._uri)

            self._auth = self._auth or asyncon.get_auth ()
            _reqtype = self._reqtype.lower ()
            rs = Dispatcher (
                self._cv, asyncon.address,
                ident = not self._mapreduce and self._get_ident () or None,
                filterfunc = self._filter, cachefs = self._cachefs,
                callback = self._collect
            )
            self._requests [rs] = asyncon
            args = (params, self._headers, self._auth, self._logger, self._meta)
            try:
                if _reqtype in ("ws", "wss"):
                    handler = ws_request_handler.RequestHandler
                    request = ws_request.Request (self._uri, *args)
                else:
                    if not self._use_cache:
                        self._add_header ("Cache-Control", "no-cache")
                    handler = http_request_handler.RequestHandler
                    try:
                        class_ = self.proto_map [_reqtype]
                    except KeyError:
                        if _reqtype == "upload":
                            request = http_request.HTTPMultipartRequest (self._uri, _reqtype, *args)
                        else:
                            request = http_request.HTTPRequest (self._uri, _reqtype, *args)
                    else:
                        request = class_ (self._uri, method, *args)
                requests += self._handle_request (request, rs, asyncon, handler)

            except:
                self._logger ("Request Creating Failed", "fail")
                self._logger.trace ()
                rs.request_failed ()
                asyncon.set_active (False)
                continue

        if requests:
            self._request = request # sample for unitest
            trigger.wakeup ()

        if _reqtype [-3:] == "rpc":
            return self

    def _avails (self):
        return len (self._nodes)

    def _get_connection (self, id = None):
        if id is None: id = self._nodes.pop ()
        else: self._nodes = []
        asyncon = self._cluster.get (id)
        self._setup (asyncon)
        return asyncon

    def _setup (self, asyncon):
        asyncon.set_timeout (self._timeout)
        if self._cv is None:
            self._cv = asyncon._cv

    def _cancel (self):
        with self._cv:
            self._canceled = True

    def _count (self):
        with self._cv:
            return len (self._requests)

    #---------------------------------------------------------
    def _fail_log (self, status):
        if self._origin:
            self._logger ("backend status is {}, {} at {} LINE {}: {}".format (
                status, self._origin [3], self._origin [1], self._origin [2], self._origin [4][0].strip ()
            ), "debug")

    def _collect (self, rs, failed = False):
        with self._cv:
            if not failed and self._canceled:
                return
            try:
                asyncon = self._requests.pop (rs)
            except KeyError:
                return

        status = rs.get_status ()
        if status == REQFAIL:
            with self._cv:
                self._results.append (rs)
            self._cluster.report (asyncon, True) # not asyncons' Fault

        elif status == TIMEOUT:
            with self._cv:
                self._results.append (rs)
            self._cluster.report (asyncon, False) # not asyncons' Fault

        elif not self._mapreduce and status == NETERR and self._retry < (self._numnodes - 1):
            self._logger ("cluster response error, switch to another...", "fail")
            self._cluster.report (asyncon, False) # exception occured
            with self._cv:
                self._retry += 1
                self._canceled = False
                self._nodes = [None]
            return self.rerequest ()

        elif status >= NETERR:
            with self._cv:
                self._results.append (rs)
            if status == NETERR:
                self._cluster.report (asyncon, False) # exception occured
            else:
                self._cluster.report (asyncon, True) # well-functioning
                rs.do_filter ()

        with self._cv:
            requests = self._requests
            callback, self._callback = self._callback, None

        if not requests:
            if callback:
                self._do_callback (callback)
            elif not failed:
                cv = self._ccv is not None and self._ccv or self._cv
                with cv:
                    cv.notify_all ()

    def _do_callback (self, callback):
        result = self.dispatch (wait = False)
        tuple_cb (result, callback)

    #-----------------------------------------------------------------
    def rerequest (self):
        self._build_request (*self._cached_request_args)

    def reset_timeout (self, timeout, ccv = None):
        with self._cv:
            self._timeout = timeout
            self._ccv = ccv
            asyncons = list (self._requests.values ())

        if timeout > 0:
            for asyncon in asyncons:
                asyncon.set_timeout (timeout)

    def set_callback (self, callback, reqid = None, timeout = None):
        if reqid is not None:
            self._meta ["__reqid"] = reqid

        if self._cv:
            with self._cv:
                requests = self._requests
                self._callback = callback
        else:
            # already finished or will use cache
            requests = self._requests
            self._callback = callback

        if not requests:
            return self._do_callback (callback)
        timeout and self.reset_timeout (timeout)

    # synchronous methods ----------------------------------------------
    def _wait (self, timeout = None):
        timeout and self.reset_timeout (timeout)
        remain = self._timeout - (time.time () - self._init_time)
        if remain > 0:
            with self._cv:
                if self._requests and not self._canceled:
                    self._cv.wait (remain)

        self._canceled = True
        requests = list (self._requests.items ())
        for rs, asyncon in requests:
            rs.set_status (TIMEOUT)
            asyncon.handle_abort () # abort imme
            self._collect (rs, failed = True)

    def dispatch (self, cache = None, cache_if = (200,), timeout = None, wait = True, reraise = False):
        if self._cached_result is not None:
            return self._cached_result
        wait and self._wait (timeout)
        rss = [rs.get_result () for rs in self._results]
        for rs in rss:
            if rs.status == NORMAL and rs.status_code < 300:
                continue
            self._fail_log (rs.status)
            reraise and rs.reraise ()

        if self._mapreduce:
            self._cached_result = Results (rss, ident = self._get_ident ())
        else:
            self._cached_result = rss [0]
        self.cache (cache, cache_if)
        return self._cached_result

    def dispatch_or_throw (self, cache = None, cache_if = (200,), timeout = None):
        return self.dispatch (cache, cache_if, reraise = True, timeout = timeout)

    def none_or_dispatch (self, cache = None, cache_if = (200,), timeout = None):
        r = self.dispatch (cache, cache_if, reraise = True, timeout = timeout)
        if r.data is not None:
          return r

    def wait (self, timeout = None, reraise = False):
        return self.dispatch (reraise = reraise, timeout = timeout)

    # direct access to data ----------------------------------------------

    def commit (self, timeout = None):
        return self.wait (timeout, True)
    wait_or_throw = commit

    def fetch (self, cache = None, cache_if = (200,), timeout = None):
        res = self._cached_result or self.dispatch (timeout = timeout, reraise = True)
        return res.fetch (cache or self._cache_timeout, cache_if)

    def one (self, cache = None, cache_if = (200,), timeout = None):
        try:
            res = self._cached_result or self.dispatch (timeout = timeout, reraise = True)
        except (PGIntegrityError, sqlite3.IntegrityError):
            raise exceptions.HTTPError ("409 Conflict")
        return res.one (cache or self._cache_timeout, cache_if)

    def then (self, func):
        from ..tasks import Future
        return Future (self, self._timeout, **self._meta).then (func)

    def cache (self, cache = 60, cache_if = (200,)):
        cache = cache or self._cache_timeout
        if not cache:
            return self
        if self._cached_result is None:
            raise ValueError("call dispatch first")
        self._cached_result.cache (cache, cache_if)
        return self
    getwait = getswait = dispatch # lower ver compat.
    getwait_or_throw = getswait_or_throw = dispatch_or_throw # lower ver compat.


# cluster base call ---------------------------------------
class _Method:
    def __init__(self, send, name):
        self.__send = send
        self.__name = name

    def __getattr__(self, name):
        return _Method(self.__send, "%s.%s" % (self.__name, name))

    def __call__(self, *args):
        return self.__send(self.__name, args)


class Proxy:
    def __init__ (self, __class, *args, **kargs):
        self.__class = __class
        self.__args = args
        self.__kargs = kargs

    def __enter__ (self):
        return self

    def __exit__ (self, type, value, tb):
        pass

    def __getattr__ (self, name):
        return _Method (self.__request, name)

    def __request (self, method, params):
        cdc = self.__class (*self.__args, **self.__kargs)
        cdc._build_request (method, params)
        return cdc


class TaskCreator:
    def __init__ (self, cluster, logger, cachesfs):
        self.cluster = cluster
        self.logger = logger
        self.cachesfs = cachesfs

    def __getattr__ (self, name):
        return getattr (self.cluster, name)

    def Server (self, uri, params = None, reqtype="rpc", headers = None, auth = None, meta = None, use_cache = True, mapreduce = False, filter = None, callback = None, cache = None, timeout = DEFAULT_TIMEOUT, caller = None):
        if type (headers) is list:
            h = {}
            for n, v in headers:
                h [n] = v
            headers = h

        if reqtype.endswith ("rpc"):
            return Proxy (Task, self.cluster, uri, params, reqtype, headers, auth, meta, use_cache, mapreduce, filter, callback, cache, timeout, caller, self.cachesfs, self.logger)
        else:
            return Task (self.cluster, uri, params, reqtype, headers, auth, meta, use_cache, mapreduce, filter, callback, cache, timeout, caller, self.cachesfs, self.logger)



#----------------------------------------------------------
# Implemetation from:
#   https://blog.grijjy.com/2018/08/29/creating-high-performance-udp-servers-on-windows-and-linux/
# Initialized at: Oct 28, 2019
# Author: Hans Roh
#----------------------------------------------------------

from . import http_server, https_server, http_request
import socket
from rs4 import asyncore, asynchat
import os, sys, errno
import time
from .lifetime import tick_timer

class http3_channel (https_server.https_channel, http_server.http_channel):
    def __init__ (self, server, data, addr):
        http_server.http_channel.__init__(self, server, None, addr)
        self.initial_data = data
        self.protocol = None # quic

        self._timer_at = None
        self._timer_id = None
        self.create_handler ()
        self.set_terminator (None)
        self.create_socket (socket.AF_INET, socket.SOCK_DGRAM)
        self.set_reuse_addr ()
        self.bind (self.server.addr)
        self.addr = addr # bind change addr, so recovering
        self.connect (self.addr)

    def writable (self):
        return self._writable_with_protocol ()

    def handle_write (self):
        written = self._handle_write_with_protocol ()
        if written and self.protocol:
            timer_at = self.protocol.get_timer()
            if self._timer_id is not None and self._timer_at != timer_at:
                tick_timer.cancel (self._timer_id)
                self._timer_id = None
            if self._timer_id is None and timer_at is not None:
                self._timer_id = tick_timer.at (timer_at, self.handle_timer)
            self._timer_at = timer_at

    def handle_timer (self):
        if not self.current_request:
            return
        now = max (self._timer_at, time.monotonic ())
        self._timer_id = None
        self._timer_at = None
        self.protocol.handle_timer (now = now)
        self.current_request.process_quic_events ()
        self.handle_write ()

    def readable (self):
        return self.connected

    def recv (self, buffer_size):
        try:
            return http_server.http_channel.recv (self, buffer_size)
        except (ConnectionRefusedError, ConnectionResetError):
            self.handle_close ()
            return b''

    def send (self, data):
        try:
            return http_server.http_channel.send (self, data)
        except (ConnectionRefusedError, ConnectionResetError):
            self.handle_close ()
            return 0

    # packet handler -----------------------------------------------------------
    def create_handler (self):
        r = http_request.http_request (self, "QUiC / HTTP/3.0", "QUiC", "/", "3.0", [])
        self.set_timeout (self.network_timeout)
        self.request_counter.inc()
        self.server.total_requests.inc ()

        for h in self.server.handlers:
            if h.match (r):
                try:
                    h.handle_request (r) # will set self.current_request
                except:
                    self.handle_error ()
                    return

    def handle_connect (self):
        self.collect_incoming_data (self.initial_data) # collect initial data
        self.initial_data = None

    def collect_incoming_data (self, data):
        self.current_request.collect_incoming_data (data)
        if self.protocol is None:
            self.protocol = self.current_request.quic


class SessionTicketStore:
    def __init__(self):
        self.tickets = {}

    def add(self, ticket):
        self.tickets [ticket.ticket] = ticket

    def pop(self, label):
        return self.tickets.pop (label, None)


class http3_server (http_server.http_server):
    ac_in_buffer_size = 65536
    sock_type = socket.SOCK_DGRAM
    VERSION = 'h3-23'

    def __init__ (self, ip, port, ssl_port, ctx, server_logger = None, request_logger = None):
        http_server.http_server.__init__ (self, ip, port, server_logger, request_logger)
        self.ssl_port = ssl_port
        self.ctx = ctx
        self.ticket_store = SessionTicketStore ()

    def _serve (self, shutdown_phase = 2):
        self.shutdown_phase = shutdown_phase

    def readable (self):
        return True

    def recv (self, buffer_size):
        return self.socket.recvfrom (buffer_size)

    def handle_read (self):
        ret = self.recv (self.ac_in_buffer_size)
        if not ret:
            return
        data, addr = ret
        if data:
            http3_channel (self, data, addr)


def init_context (certfile, keyfile, pass_phrase):
    from aioquic.h3.connection import H3_ALPN
    from aioquic.quic.configuration import QuicConfiguration
    import ssl

    ctx = QuicConfiguration (alpn_protocols = H3_ALPN, is_client = False)
    ctx.load_cert_chain (certfile, keyfile, pass_phrase)
    ctx.verify_mode = ssl.CERT_NONE
    return ctx




import os, sys, re, types, time
from rs4  import pathtool, importer, evbus
import threading
from types import FunctionType as function
import copy
from .. import lifetime
import inspect
from rs4 import attrdict
from importlib import reload

def set_default (cf):
	cf.max_post_body_size = 1 * 1024 * 1024
	cf.max_cache_size = 1 * 1024 * 1024
	cf.max_multipart_body_size = 20 * 1024 * 1024
	cf.max_upload_file_size = 20 * 1024 * 1024

def Config (preset = False):
	cf = attrdict.AttrDict ()
	if preset:
		set_default (cf)
	return cf


class Module:
	def __init__ (self, wasc, handler, bus, route, directory, libpath, pref = None):
		self.wasc = wasc
		self.bus = bus
		self.handler = handler
		self.pref = pref
		self.last_reloaded = time.time ()
		self.app = None
		self.django = False
		self.set_route (route)
		self.directory = directory
		self.has_life_cycle = False

		if type (libpath) is str:
			try:
				libpath, self.appname = libpath.split (":", 1)
			except ValueError:
				libpath, self.appname = libpath, "app"
			self.libpath = libpath
			self.script_name = "%s.py" % libpath
			self.module, self.abspath = importer.importer (directory, libpath)
			self.start_app ()

		else:
			# libpath is app object, might be added by unittest
			self.appname, self.module, self.abspath = "app", None, os.path.join (directory, 'dummy')
			self.app = libpath
			self.app.use_reloader = False
			self.start_app ()

		app = self.app or getattr (self.module, self.appname)
		self.has_life_cycle and app.life_cycle ("before_mount", self.wasc)

	def __repr__ (self):
		return "<Module routing to %s at %x>" % (self.route, id(self))

	def get_callable (self):
		return self.app or getattr (self.module, self.appname)

	def start_app (self, reloded = False):
		func = None
		app = self.app or getattr (self.module, self.appname)

		if hasattr (app, "set_logger"):
			app.set_logger (self.wasc.logger.get ("app"))
		self.django = str (app.__class__).find ("django.") != -1
		self.has_life_cycle = hasattr (app, "life_cycle")

		if self.pref:
			for k, v in copy.copy (self.pref).items ():
				if k == "config":
					if not hasattr (app, 'config'):
						app.config = v
					else:
						for k, v in copy.copy (self.pref.config).items ():
							app.config [k] = v
				else:
					setattr (app, k, v)

		if hasattr (app, "_aliases"):
			while app._aliases:
				self.wasc.add_cluster (*app._aliases.pop (0))

		if not hasattr (app, "config"):
			app.config = Config (True)

		elif not hasattr (app.config, "max_post_body_size"):
			set_default (app.config)

		if hasattr (app, "mountables"):
			for _args, _karg in app.mountables:
				app.mount (*_args, **_karg)

		if hasattr (app, "max_client_body_size"):
			val = app.max_client_body_size
			app.config.max_post_body_size = val
			app.config.max_multipart_body_size = val
			app.config.max_upload_file_size = val

		if hasattr (app, "set_home"):
			app.set_home (os.path.dirname (self.abspath), self.module)

		if hasattr (app, "commit_events_to"):
			app.commit_events_to (self.bus)

		self.set_devel_env ()
		self.update_file_info ()

		try:
			if not reloded:
				func = app.start
			else:
				func = app.restart
		except AttributeError:
			pass

		if func:
			# temporary current handler
			self.wasc.handler = self.handler
			func (self.wasc, self.route)
			self.wasc.handler = None

	def mounted (self):
		app = self.app or getattr (self.module, self.appname)
		self.has_life_cycle and app.life_cycle ("mounted", self.wasc ())
		self.has_life_cycle and app.life_cycle ("mounted_or_reloaded", self.wasc ())

	def umounted (self):
		app = self.app or getattr (self.module, self.appname)
		self.has_life_cycle and app.life_cycle ("umounted", self.wasc)

	def cleanup (self):
		app = self.app or getattr (self.module, self.appname)
		self.has_life_cycle and app.life_cycle ("before_umount", self.wasc ())
		try: app.cleanup ()
		except AttributeError: pass

	def check_django_reloader (self, now):
		from django.utils import autoreload

		changed = autoreload.code_changed ()
		if changed:
			self.wasc.logger ("app", "reloading app, %s" % self.abspath, "debug")
			self.last_reloaded = time.time ()
			lifetime.shutdown (3, 0)

	def set_devel_env (self):
		self.debug = False
		self.use_reloader = False
		app = self.app or getattr (self.module, self.appname)

		if os.environ.get ("SKITAI_ENV") == "PRODUCTION":
			app.debug = False
			app.use_reloader = False
			return

		try: self.debug = app.debug
		except AttributeError: pass
		try: self.use_reloader = app.use_reloader
		except AttributeError: pass
		else:
			if self.use_reloader and self.django:
				lifetime.maintern.sched (1.0, self.check_django_reloader)

	def update_file_info (self):
		if self.module is None:
			# app directly mounted
			return
		stat = os.stat (self.abspath)
		self.file_info = (stat.st_mtime, stat.st_size)

	def maybe_reload (self):
		if not self.use_reloader:
			return

		if self.django:
			# see check_django_reloader
			return

		if time.time () - self.last_reloaded < 1.0:
			return

		stat = os.stat (self.abspath)
		if self.file_info != (stat.st_mtime, stat.st_size):
			oldapp = getattr (self.module, self.appname)
			self.has_life_cycle and oldapp.life_cycle ("before_reload", self.wasc ())

			try:
				reloaded = importer.reimporter (self.module, self.directory, self.libpath)
			except:
				self.module.app = oldapp
				raise
			else:
				if not reloaded:
					return
				self.module, self.abspath = reloaded
				if hasattr (oldapp, "remove_events"):
					oldapp.remove_events (self.bus)
				PRESERVED = []
				if hasattr (oldapp, "PRESERVES_ON_RELOAD"):
					PRESERVED = [(attr, getattr (oldapp, attr)) for attr in oldapp.PRESERVES_ON_RELOAD]

				self.start_app (reloded = True)
				newapp = getattr (self.module, self.appname)
				for attr, value in PRESERVED:
					setattr (newapp, attr, value)

				# reloaded
				self.has_life_cycle and newapp.life_cycle ("reloaded", self.wasc ())
				self.has_life_cycle and newapp.life_cycle ("mounted_or_reloaded", self.wasc ())
				self.last_reloaded = time.time ()
				self.wasc.logger ("app", "reloading app, %s" % self.abspath, "debug")

	def set_route (self, route):
		route = route
		while route and route [-1] == "/":
			route = route [:-1]
		self.route = route + "/"
		self.route_len = len (self.route) - 1

	def get_route (self):
		return self.route

	def get_path_info (self, path):
		path_info = ("/" + path) [self.route_len:]
		return path_info

	def __call__ (self, env, start_response):
		self.use_reloader and self.maybe_reload ()
		app = self.app or getattr (self.module, self.appname)
		return app (env, start_response)


class ModuleManager:
	modules = {}
	def __init__(self, wasc, handler):
		self.wasc = wasc
		self.handler = handler
		self.modules = {}
		self.modnames = {}
		self.bus = evbus.EventBus ()
		self.cc = 0

	def __getitem__ (self, name):
		return self.modnames [name].get_callable ()

	def build_url (self, thing, *args, **kargs):
		a, b = thing.split (":", 1)
		return self.modnames [a].get_callable ().build_url (b, *args, **kargs)

	def add_module (self, route, directory, modname, pref):
		dispname = modname
		if type (modname) is str and  modname.startswith ("__export__"):
			dispname = "{}:app".format (os.path.basename (os.path.dirname (os.path.dirname (directory))))

		if dispname in self.modnames:
			self.wasc.logger ("app", "app file name collision detected '%s'" % dispname, "error")
			return

		if not route:
			route = "/"
		elif not route.endswith ("/"):
			route = route + "/"

		try:
			module = Module (self.wasc, self.handler, self.bus, route, directory, modname, pref)
		except:
			self.wasc.logger.trace ("app")
			self.wasc.logger ("app", "[error] app load failed: %s" % dispname)
		else:
			self.wasc.logger ("app", "[info] app %s mounted" % dispname)
			if route in self.modules:
				self.wasc.logger ("app", "[info] app route collision detected: %s at %s <-> %s" % (route, module.abspath, self.modules [route].abspath), "warn")
			self.modules [route] = module
			if type (dispname) is str:
				# possibley direct app object
				self.modnames [dispname.split (":", 1)[0]] = module

	def get_app (self, script_name):
		route = self.has_route (script_name)
		if route in (0, 1): # 404, 301
			return None

		try:
			app = self.modules [route]
		except KeyError:
			return None
		return app

	def has_route (self, script_name):
		# 0: 404
		# 1: 301 => /skitai => /skitai/
		if script_name == "/" and "/" in self.modules:
			return "/"

		cands = []
		for route in self.modules:
			if script_name == route:
				cands.append (route)
			elif script_name + "/" == route:
				return 1
			elif script_name.startswith (route [-1] != "/" and route + "/" or route):
				cands.append (route)

		if cands:
			if len (cands) == 1:
				return cands [0]
			cands.sort (key = lambda x: len (x))
			return cands [-1]

		elif "/" in self.modules:
			return "/"

		return 0

	def unload (self, route):
		module = self.modules [route]
		self.wasc.logger ("app", "[info] unloading app: %s" % route)
		try:
			self.wasc.logger ("app", "[info] ..cleanup app: %s" % route)
			module.cleanup ()
		except AttributeError:
			pass
		except:
			self.wasc.logger.trace ("app")
		del module
		del self.modules [route]

	def cleanup (self):
		self.wasc.logger ("app", "[info] cleanup apps")
		for route, module in list(self.modules.items ()):
			try:
				self.wasc.logger ("app", "[info] ..cleanup app: %s" % route)
				module.cleanup ()
			except AttributeError:
				pass
			except:
				self.wasc.logger.trace ("app")

	def status (self):
		d = {}
		for path, module in list(self.modules.items ()):
			d ['<a href="%s">%s</a>' % (path, path)] = module.abspath
		return d





from ..httpbase import cluster_manager
from aquests.dbapi import asynpsycopg2, synsqlite3, asynredis, asynmongo, syndbi, dbpool
from skitai import DB_PGSQL, DB_SQLITE3, DB_REDIS, DB_MONGODB, DB_SYN_PGSQL, DB_SYN_REDIS, DB_SYN_MONGODB
from . import endpoints
from sqlphile import pg2, db3
            
class ClusterManager (cluster_manager.ClusterManager):
    backend_keep_alive = 10
    backend = True
    class_map = dbpool.DBPool.class_map
    def __init__ (self, name, cluster, dbtype = DB_PGSQL, access = [], max_conns = 200, logger = None):
        self.dbtype = dbtype
        self._cache = []
        cluster_manager.ClusterManager.__init__ (self, name, cluster, 0, access, max_conns, logger)

    @classmethod
    def add_class (cls, name, class_):
        cls.class_map [name] = class_

    def match (self, request):
        return False # not serverd by url
    
    def create_asyncon (self, member):        
        if self.dbtype == DB_SQLITE3:
            asyncon = self.class_map [DB_SQLITE3] (member, None, self.lock, self.logger)
            nodeid = member
            self._cache.append (((member, 0), "", ("", "")))
        
        else:
            if member.find ("@") != -1:
                auth, netloc = self.parse_member (member)
                try:
                    server, db = netloc.split ("/", 1)
                except ValueError:
                    server, db = netloc, ""
                    
            else:                
                db, user, passwd = "", "", ""
                args = member.split ("/", 3)
                if len (args) == 4:     server, db, user, passwd = args
                elif len (args) == 3:     server, db, user = args
                elif len (args) == 2:     server, db = args        
                else: server = args [0]
                auth = (user, passwd)
                
            try: 
                host, port = server.split (":", 1)
                server = (host, int (port))
            except ValueError: 
                server = (server, 5432)
            
            try:
                conn_class = self.class_map [self.dbtype]
            except KeyError:
                raise TypeError ("Unknown DB type: %s" % self.dbtype)
            
            asyncon = conn_class (server, (db, auth), self.lock, self.logger)    
            self.backend and asyncon.set_backend (self.backend_keep_alive)            
            nodeid = server
            self._cache.append ((server, db, auth))
                
        return nodeid, asyncon # nodeid, asyncon
    
    def get_endpoints (self):
        return endpoints.make_endpoints (self.dbtype, self._cache)
    
    def _openv (self, wrap):
        # transaction mode
        if self.dbtype in (DB_PGSQL, DB_SYN_PGSQL):
            conn = endpoints.make_endpoints (self.dbtype, [self._cache [0]]) [0]
            if wrap == "open2":
                return pg2.open2 (conn)
            return pg2.open3 (conn)
        elif self.dbtype == DB_SQLITE3:            
            conn = endpoints.make_endpoints (self.dbtype, [self._cache [0]]) [0]
            if wrap == "open2":
                return db3.open2 (conn)
            return db3.open3 (conn)
        raise TypeError ("Only DB_PGSQL or DB_SQLITE3")    

    def open2 (self):
        # single conn, single cursor
        return self._openv ('open2')

    def open3 (self):
        # single connection, multi cursors
        # it is replacable with was.db
        return self._openv ('open3')


from . import wsgi_handler
import skitai
from rs4 import producers
from h2.connection import H2Connection, GoAwayFrame, DataFrame, H2Configuration
from h2.exceptions import ProtocolError, NoSuchStreamError, StreamClosedError, FlowControlError
from h2.events import TrailersReceived, DataReceived, RequestReceived, StreamEnded, PriorityUpdated, ConnectionTerminated, StreamReset, WindowUpdated, RemoteSettingsChanged
from h2.settings import SettingCodes
from h2.errors import ErrorCodes
from .http2.request import request as http2_request
from .http2.vchannel import fake_channel, data_channel
from aquests.protocols.http2.request_handler import FlowControlWindow
import threading
from io import BytesIO
import time
import sys

class http2_producer:
    SIZE_BUFFER = 16384
    def __init__ (self, conn, lock, stream_id, headers, producer, trailers, force_close, depends_on = 0, priority = 1, log_hook = None):
        self.conn = conn
        self.lock = lock
        self.stream_id = stream_id
        self.headers = headers
        self.producer = producer
        self.trailers = trailers
        self.force_close = force_close
        self.depends_on = depends_on
        self.priority = priority
        self.log_hook = log_hook
        self.bytes = 0
        self.next_data = None
        self._end_stream = False
        self._last_sent = time.time ()

    def __lt__ (self, other):
        if self.depends_on == other.depends_on:
            return self.priority > other.priority # descending
        return self.depends_on < other.depends_on # ascending

    def set_stream_ended (self):
        self._end_stream = True

    def is_stream_ended (self):
        return self._end_stream

    def log (self):
        self.log_hook and self.log_hook (self.bytes)

    def send_final_data (self, data):
        with self.lock:
            self.conn.send_data (self.stream_id, data, end_stream = not self.trailers)
            if self.trailers:
                self.conn.send_headers (self.stream_id, self.trailers, end_stream = True)
        self.log ()

    def local_flow_control_window (self):
        lfcw = self.conn.local_flow_control_window (self.stream_id)
        if lfcw == 0:
            # flow control error, graceful close
            if time.time () - self._last_sent > 10:
                raise FlowControlError
        return lfcw

    def produce (self):
        if self.is_stream_ended ():
            return 0
        lfcw = self.local_flow_control_window ()
        if not lfcw:
            return 0

        self._last_sent = time.time ()
        sent = 0
        if self.headers:
            with self.lock:
                self.conn.send_headers (self.stream_id, self.headers, end_stream = not self.producer and not self.trailers)
            self.headers = None
            sent = 1

        if not self.producer:
            self.set_stream_ended ()
            self.log ()
            return sent

        if self.next_data:
            data, self.next_data = self.next_data, None
        else:
            if hasattr (self.producer, 'ready') and not self.producer.ready ():
                return sent
            data = self.producer.more ()
            if not data:
                self.set_stream_ended ()
                self.send_final_data (b'')
                return 1

        avail_data_length = min (self.SIZE_BUFFER, lfcw)
        if len (data) > avail_data_length:
            data, self.next_data = data [:avail_data_length], data [avail_data_length:]

        self.bytes += len (data)
        if self.next_data is None and (not hasattr (self.producer, 'ready') or self.producer.ready ()):
            self.next_data = self.producer.more ()
            if not self.next_data:
                self.set_stream_ended ()

        if self.is_stream_ended ():
            self.send_final_data (data)
        else:
            with self.lock:
                self.conn.send_data (self.stream_id, data, end_stream = False)
        return 1


class http2_request_handler (FlowControlWindow):
    collector = None
    producer = None
    http11_terminator = 24
    producer_class = http2_producer
    altsvc = None

    def __init__ (self, handler, request):
        self.handler = handler
        self.wasc = handler.wasc
        self.request = request
        self.channel = request.channel
        self.conn = H2Connection (H2Configuration (client_side = False))

        self.frame_buf = self.conn.incoming_buffer
        self.frame_buf.max_frame_size = self.conn.max_inbound_frame_size
        self.data_length = 0
        self.current_frame = None
        self.rfile = BytesIO ()
        self.buf = b""
        self._got_preamble = False
        self.default_varialbes ()
        self.altsvc = request.channel.server.altsvc

    def default_varialbes (self):
        self.producers = []
        self.requests = {}
        self.priorities = {}
        self._has_sendables = False
        self._closed = False
        self._plock = threading.Lock () # for self.conn
        self._clock = threading.Lock () # for self.x

    def send_data (self):
        self._has_sendables = True

    def has_sendables (self):
        return self._has_sendables

    def data_from_producers (self, force_close_error_code):
        for i in range (len (self.producers)):
            with self._clock:
                producer = self.producers [0]
            produced = producer.produce ()
            if producer.is_stream_ended ():
                with self._clock:
                    self.producers.pop (0)
                r = self.get_request (producer.stream_id)
                self.remove_request (producer.stream_id)
                if producer.force_close:
                    return self.go_away (force_close_error_code)
            if produced:
                break
            with self._clock:
                self.producers.append (self.producers.pop (0))
            continue

    def data_to_send (self):
        self.data_from_producers (ErrorCodes.CANCEL)
        with self._plock:
            data_to_send = self.conn.data_to_send ()
        if not data_to_send and not self.producers:
            self._has_sendables = False
        if data_to_send:
            return [data_to_send]
        return []

    def go_away (self, errcode = 0, msg = None):
        with self._plock:
            self.conn.close_connection (errcode, msg)
        if self.channel:
            self.send_data ()
            self.channel.close_when_done ()

    def close (self, errcode = 0, msg = None):
        if self._closed: return
        self._closed = True
        with self._clock:
            stream_ids = list (self.requests.keys ())
        for stream_id in stream_ids:
            self.remove_request (stream_id)
        self.go_away ()

    def enter_shutdown_process (self):
        self.close (ErrorCodes.NO_ERROR)

    def closed (self):
        return self._closed

    def handle_preamble (self):
        if self.request.version.startswith ("2."):
            self.channel.set_terminator (6) # SM\r\n\r\n

    def initiate_connection (self):
        self.handle_preamble ()
        h2settings = self.request.get_header ("HTTP2-Settings")
        if h2settings:
            self.conn.initiate_upgrade_connection (h2settings)
        else:
            self.conn.initiate_connection()
        self.send_data ()
        if self.request.version == "1.1":
            self.handle_request (1, self.upgrade_header ())

    def upgrade_header (self):
        headers = [
            (":method", self.request.command.upper ()),
            (":path", self.request.uri),
        ]
        for line in self.request.get_header ():
            try:
                k, v = line.split (": ", 1)
            except ValueError:
                k, v = line.split (":", 1)
            k = k.lower ()
            if k in ("http2-settings", "connection", "upgrade"):
                continue
            headers.append ((k, v))
        return headers

    def collect_incoming_data (self, data):
        #print ("RECV", repr (data), len (data), '::', self.channel.get_terminator ())
        if not data:
            # closed connection
            self.close (True)
            return

        if self.data_length:
            self.rfile.write (data)
        else:
            self.buf += data

    def set_frame_data (self, data):
        if not self.current_frame:
            return []
        self.current_frame.parse_body (memoryview (data))
        self.current_frame = self.frame_buf._update_header_buffer (self.current_frame)
        with self._plock:
            events = self.conn._receive_frame (self.current_frame)
        return events

    def set_terminator (self, terminator):
        self.channel.set_terminator (terminator)

    def found_terminator (self):
        buf, self.buf = self.buf, b""

        #print ("FOUND", repr (buf), '::', self.data_length, self.channel.get_terminator ())
        #print ("FOUND", self.request.version, self.request.command, self.data_length)
        events = None
        if self.request.version == "1.1" and self.data_length:
            self.request.version = "2.0" # upgrade
            data = self.rfile.getvalue ()
            with self._clock:
                r = self.requests [1]
            r.channel.set_data (data, len (data))
            r.set_stream_ended ()
            self.data_length = 0
            self.rfile.seek (0)
            self.rfile.truncate ()
            self.channel.set_terminator (24) # for premble

        elif not self._got_preamble:
            if not buf.endswith (b"SM\r\n\r\n"):
                raise ProtocolError ("Invalid preamble")
            self._got_preamble = True
            self.channel.set_terminator (9)

        elif self.data_length:
            events = self.set_frame_data (self.rfile.getvalue ())
            self.current_frame = None
            self.data_length = 0
            self.rfile.seek (0)
            self.rfile.truncate ()
            self.channel.set_terminator (9) # for frame header

        elif buf:
            self.current_frame, self.data_length = self.frame_buf._parse_frame_header (buf)
            self.frame_buf.max_frame_size = self.data_length

            if self.data_length == 0:
                events = self.set_frame_data (b'')
            self.channel.set_terminator (self.data_length == 0 and 9 or self.data_length)    # next frame header

        else:
            raise ProtocolError ("Frame decode error")

        if events:
            self.handle_events (events)

    def pushable (self):
        return self.conn.remote_settings [SettingCodes.ENABLE_PUSH]

    def push_promise (self, stream_id, request_headers, addtional_request_headers):
        promise_stream_id = self.conn.get_next_available_stream_id ()
        headers = request_headers + addtional_request_headers
        with self._plock:
            self.conn.push_stream (stream_id, promise_stream_id, headers)
        event = RequestReceived ()
        event.stream_id = promise_stream_id
        event.headers = headers
        self.handle_events ([event])

    def handle_response (self, stream_id, headers, trailers, producer, do_optimize, force_close = False):
        with self._clock:
            try:
                depends_on, weight = self.priorities [stream_id]
            except KeyError:
                depends_on, weight = 0, 1
            else:
                del self.priorities [stream_id]

        if self.altsvc:
            headers.append (("alt-svc", '{}=":{}"; ma=86400'.format (self.altsvc.VERSION, self.altsvc.ssl_port)))

        if trailers:
            assert producer, "http/2 or 3's trailser requires body"
        if producer and do_optimize:
            producer = producers.globbing_producer (producer)

        r = self.get_request (stream_id)
        if r:
            self.producers.append (self.producer_class (self.conn, self._plock, stream_id, headers, producer, trailers, force_close, depends_on, weight, r.response.maybe_log))
            self.producers.sort ()
            self.send_data ()

    def remove_request (self, stream_id):
        r = self.get_request (stream_id)
        if not r: return
        with self._clock:
            try: del self.requests [stream_id]
            except KeyError: pass
        r.protocol = None

    def get_request (self, stream_id):
        r = None
        with self._clock:
            try: r =    self.requests [stream_id]
            except KeyError: pass
        return r

    def adjust_flow_control_window (self, stream_id):
        if self.conn.inbound_flow_control_window < self.MIN_IBFCW:
            with self._clock:
                self.conn.increment_flow_control_window (1048576)

        rfcw = self.conn.remote_flow_control_window (stream_id)
        if rfcw < self.MIN_RFCW:
            try:
                with self._clock:
                    self.conn.increment_flow_control_window (1048576, stream_id)
            except StreamClosedError:
                pass

    def handle_events (self, events):
        for event in events:
            if isinstance(event, RequestReceived):
                self.handle_request (event.stream_id, event.headers, has_data_frame = not event.stream_ended)

            elif isinstance(event, TrailersReceived):
                self.handle_trailers (event.stream_id, event.headers)

            elif isinstance(event, StreamReset):
                if event.remote_reset:
                    r = self.get_request (event.stream_id)
                    if r and r.collector:
                        try: r.collector.stream_has_been_reset ()
                        except AttributeError: pass
                        r.set_stream_ended ()
                    self.channel.producer_fifo.remove (event.stream_id)

            elif isinstance(event, ConnectionTerminated):
                self.close ()

            elif isinstance(event, PriorityUpdated):
                if event.exclusive:
                    # rebuild depend_ons
                    for stream_id in list (self.priorities.keys ()):
                        depends_on, weight = self.priorities [stream_id]
                        if depends_on == event.depends_on:
                            self.priorities [stream_id] = [event.stream_id, weight]

                with self._clock:
                    self.priorities [event.stream_id] = [event.depends_on, event.weight]

            elif isinstance(event, DataReceived):
                r = self.get_request (event.stream_id)
                if not r:
                    self.go_away (ErrorCodes.PROTOCOL_ERROR)
                else:
                    try:
                        r.channel.set_data (event.data, event.flow_controlled_length)
                    except ValueError:
                        # from vchannel.handle_read () -> collector.collect_inconing_data ()
                        self.go_away (ErrorCodes.CANCEL)
                    else:
                        self.adjust_flow_control_window (event.stream_id)

            elif isinstance(event, StreamEnded):
                r = self.get_request (event.stream_id)
                if r:
                    if r.collector:
                        r.channel.handle_read ()
                        r.channel.found_terminator ()
                    r.set_stream_ended ()
                    if r.response.is_done ():
                        # DO NOT REMOVE before responsing:
                        # this is for async streaming request like proxy request
                        self.remove_request (event.stream_id)

        self.send_data ()

    def reset_stream (self, stream_id, errcode = ErrorCodes.CANCEL):
        closed = False
        with self._plock:
            try:
                self.conn.reset_stream (stream_id, error_code = errcode)
            except StreamClosedError:
                closed = True
        if not closed:
            self.send_data ()
        self.remove_request (stream_id)
        #self.request.logger ("stream reset (stream_id:%d, error:%d)" % (stream_id, errcode), "info")

    def handle_trailers     (self, stream_id, headers):
        r = self.get_request (stream_id)
        for k, v in headers:
            r.header.append (
                "{}: {}".format (k.decode ("utf8"), v.decode ("utf8"))
            )

    def handle_request (self, stream_id, headers, has_data_frame = False):
        #print ("++REQUEST: %d" % stream_id, headers)
        command = "GET"
        uri = "/"
        scheme = "http"
        authority = ""
        cl = 0
        h = []
        cookies = []

        if type (headers [0][0]) is bytes:
            headers = [(k.decode ("utf8"), v.decode ("utf8")) for k, v in headers]

        for k, v in headers:
            #print ('HEADER:', k, v)
            if k[0] == ":":
                if k == ":method": command = v.upper ()
                elif k == ":path": uri = v
                elif k == ":scheme": scheme = v.lower ()
                elif k == ":authority":
                    authority = v
                    if authority:
                        h.append ("host: %s" % authority.lower ())
                continue
            if k == "content-length":
                cl = int (v)
            elif k == "cookie":
                cookies.append (v)
                continue
            h.append ("%s: %s" % (k, v))

        if cookies:
            h.append ("Cookie: %s" % "; ".join (cookies))

        should_have_collector = False
        if command == "CONNECT":
            first_line = "%s %s HTTP/%s" % (command, authority, self.request.version)
            vchannel = self.channel
        else:
            first_line = "%s %s HTTP/%s" % (command, uri, self.request.version)
            if command in ("POST", "PUT", "PATCH"):
                should_have_collector = True
                if not cl and has_data_frame:
                    # dummy content-lengng for data frames
                    cl = -1
                    h.append ('content-length: {}'.format (cl))
                vchannel = data_channel (stream_id, self.channel, cl)
            else:
                if stream_id == 1:
                    self.request.version = "2.0"
                vchannel = fake_channel (stream_id, self.channel)

        r = http2_request (
            self,  scheme, stream_id,
            vchannel, first_line, command.lower (), uri, self.request.version, h
        )
        vchannel.current_request = r
        with self._clock:
            self.channel.request_counter.inc()
            self.channel.server.total_requests.inc()

        h = self.handler.default_handler
        if not h.match (r):
            try: r.response.error (404)
            except: pass
            return

        with self._clock:
            self.requests [stream_id] = r

        try:
            h.handle_request (r)

        except:
            self.channel.server.trace()
            try: r.response.error (500)
            except: pass

        else:
            if should_have_collector and cl > 0:
                if r.collector is None:
                    # POST but too large body or 3xx, 4xx
                    if stream_id == 1 and self.request.version == "1.1":
                        self.channel.close_when_done ()
                        self.remove_request (1)
                    else:
                        self.go_away (ErrorCodes.PROTOCOL_ERROR)
                else:
                    if stream_id == 1 and self.request.version == "1.1":
                        self.data_length = cl
                        self.set_terminator (cl)


class h2_request_handler (http2_request_handler):
    http11_terminator = None

    def handle_preamble (self):
        if self.request.version.startswith ("2."):
            self.conn.receive_data ("PRI * HTTP/2.0\r\n\r\n")
            self.channel.set_terminator (None)

    def collect_incoming_data (self, data):
        with self._plock:
            events = self.conn.receive_data (data)
        self.handle_events (events)


class Handler (wsgi_handler.Handler):
    keep_alive = 120

    def __init__(self, wasc, default_handler = None):
        wsgi_handler.Handler.__init__(self, wasc, None)
        self.default_handler = default_handler

    def match (self, request):
        return True

    def handle_request (self, request):
        is_http2 = False
        if request.command == "pri" and request.uri == "*" and request.version == "2.0":
            is_http2 = True
        else:
            upgrade = request.get_header ("upgrade")
            is_http2 = upgrade and upgrade.lower () == "h2c" and request.version == "1.1"

        if not is_http2:
            return self.default_handler.handle_request (request)

        http2 = http2_request_handler (self, request)
        request.channel.die_with (http2, "http2 connection")
        request.channel.set_socket_timeout (self.keep_alive)

        if request.version == "1.1":
            request.response (
                "101 Switching Protocol",
                headers = [("Connection",  "upgrade"), ("Upgrade", "h2c"), ("Server", skitai.NAME.encode ("utf8"))]
            )
            request.response.done (upgrade_to = (http2, http2.http11_terminator))
        else:
            request.channel.current_request = http2

        request.channel.link_protocol_writer ()
        http2.initiate_connection ()



from ..exceptions import HTTPError
from ..utility import make_pushables
import sys

WAS_FACTORY = None

class corequest:
    def _get_was (self):
        global WAS_FACTORY
        if WAS_FACTORY is None:
            from skitai import was
            WAS_FACTORY = was

        try:
            _was = WAS_FACTORY._get ()
        except TypeError:
            return
        else:
            assert hasattr (_was, 'app'), 'corequest future is available on only Atila'
        return _was._clone (True)

    def _late_respond (self, tasks):
        # NEED self._fulfilled and self._was
        response = self._was.response
        try:
            content = self._fulfilled (self._was, tasks)
            self._fulfilled = None
            will_be_push = make_pushables (response, content)
            content = None
        except MemoryError:
            raise
        except HTTPError as e:
            response.start_response (e.status)
            content = response.build_error_template (e.explain, e.errno, was = self._was)
        except:
            self._was.traceback ()
            response.start_response ("502 Bad Gateway")
            content = response.build_error_template (self._was.app.debug and sys.exc_info () or None, 0, was = self._was)

        if content:
           will_be_push = make_pushables (response, content)

        if will_be_push is None:
            return

        for part in will_be_push:
            if len (will_be_push) == 1 and type (part) is bytes and len (response) == 0:
                response.update ("Content-Length", len (part))
            response.push (part)
        response.done ()

    # basic methods --------------------------------------
    def get_timeout (self):
        return self._timeout

    def set_timeout (self, timeout):
        self._timeout = timeout

    def returning (self, returning):
        # coreauest.then (callback).returning ("201 Created")
        return returning

    # implementables --------------------------------------
    def then (self, func):
        # usally return self and chaing with returning ()
        raise NotImplementedError

    def cache (self, cache = 60, cache_if = (200,)):
        raise NotImplementedError

    def dispatch (self, cache = None, cache_if = (200,), timeout = None):
        # response object with data
        raise NotImplementedError

    def wait (self, timeout = None):
        # response object without data
        raise NotImplementedError

    def commit (self, timeout = None):
        # return None. if error had been occured will be raised
        raise NotImplementedError

    def fetch (self, cache = None, cache_if = (200,), timeout = None):
        # return data. if error had been occured will be raised
        raise NotImplementedError

    def one (self, cache = None, cache_if = (200,), timeout = None):
        # return data with only one element. if error had been occured will be raised
        raise NotImplementedError


class response (corequest):
    pass



from . import http2_handler
import threading
import time
import os
from . import http2_handler
from aioquic.quic import events
from aioquic.h3 import connection as h3
from aioquic.h3.events import DataReceived, HeadersReceived, PushPromiseReceived, H3Event
from aioquic.h3.exceptions import H3Error, NoAvailablePushIDError
from aioquic.quic.connection import stream_is_unidirectional
from aioquic.buffer import Buffer
from dataclasses import dataclass
import collections
from aioquic.quic.packet import (
    PACKET_TYPE_INITIAL,
    encode_quic_retry,
    encode_quic_version_negotiation,
    pull_quic_header,
)
from aioquic.quic.retry import QuicRetryTokenHandler
from aioquic.quic.connection import QuicConnection

class http3_producer (http2_handler.http2_producer):
    def local_flow_control_window (self):
        return self.SIZE_BUFFER

class http3_request_handler (http2_handler.http2_request_handler):
    producer_class = http3_producer
    stateless_retry = False
    conns = {}
    def __init__ (self, handler, request):
        self.handler = handler
        self.wasc = handler.wasc
        self.request = request
        self.channel = request.channel
        self.quic = None # QUIC protocol
        self.conn = None # HTTP3 Protocol
        self._retry = QuicRetryTokenHandler() if self.stateless_retry else None
        self.default_varialbes ()

    def data_to_send (self):
        if self.quic is None:
            return []
        self.data_from_producers (h3.ErrorCode.HTTP_REQUEST_CANCELLED)
        with self._plock:
            frames = [data for data, addr in self.quic.datagrams_to_send (now = time.monotonic ())]
        if not frames and not self.producers:
            self._has_sendables = False
        return frames

    def make_connection (self, channel, data):
        ctx = channel.server.ctx

        buf = Buffer (data=data)
        header = pull_quic_header (
            buf, host_cid_length = ctx.connection_id_length
        )
        # version negotiation
        if header.version is not None and header.version not in ctx.supported_versions:
            self.channel.push (
                encode_quic_version_negotiation (
                    source_cid = header.destination_cid,
                    destination_cid = header.source_cid,
                    supported_versions = ctx.supported_versions,
                )
            )
            return

        conn = self.conns.get (header.destination_cid)
        if conn:
            conn._linked_channel.close ()
            conn._linked_channel = channel
            self.quic = conn._quic
            self.conn = conn
            return

        if header.packet_type != PACKET_TYPE_INITIAL or len (data) < 1200:
            return

        original_connection_id = None
        if self._retry is not None:
            if not header.token:
                # create a retry token
                channel.push (
                    encode_quic_retry (
                        version = header.version,
                        source_cid = os.urandom (8),
                        destination_cid = header.source_cid,
                        original_destination_cid = header.destination_cid,
                        retry_token = self._retry.create_token (channel.addr, header.destination_cid)))
                return
            else:
                try:
                    original_connection_id = self._retry.validate_token (channel.addr, header.token)
                except ValueError:
                    return

        self.quic = QuicConnection (
            configuration = ctx,
            logger_connection_id = original_connection_id or header.destination_cid,
            original_connection_id = original_connection_id,
            session_ticket_fetcher = channel.server.ticket_store.pop,
            session_ticket_handler = channel.server.ticket_store.add
        )
        self.conn = h3.H3Connection (self.quic)
        self.conn._linked_channel = channel

    def collect_incoming_data (self, data):
        print ('collect_incoming_data', self.quic, len (data))
        if self.quic is None:
             self.make_connection (self.channel, data)
             if self.quic is None:
                 return
        with self._plock:
            self.quic.receive_datagram (data, self.channel.addr, time.monotonic ())
        self.process_quic_events ()
        self.send_data ()

    def go_away (self, errcode = h3.ErrorCode.HTTP_NO_ERROR, msg = None):
        with self._plock:
            self.quic.close (error_code=errcode, frame_type = h3.FrameType.GOAWAY, reason_phrase=msg)
        if self.channel:
            self.send_data ()
            self.channel.close_when_done ()

    def ping (self, uid):
        with self._plock:
            self.quic.send_ping (uid)
        self.send_data ()

    def pushable (self):
        return True

    def process_quic_events (self):
        with self._plock:
            event = self.quic.next_event ()
        while event is not None:
            print (event.__class__.__name__)
            if isinstance(event, events.ConnectionIdIssued):
                self.conns [event.connection_id] = self.conn

            elif isinstance(event, events.ConnectionIdRetired):
                assert self.conns [event.connection_id] == self.conn
                conn = self.conns.pop (event.connection_id)
                conn._linked_channel = None

            elif isinstance(event, events.ConnectionTerminated):
                for cid, conn in list (self.conns.items()):
                    if conn == self.conn:
                        conn._linked_channel = None
                        del self.conns [cid]
                self.close ()

            elif isinstance(event, events.HandshakeCompleted):
                pass

            elif isinstance(event, events.PingAcknowledged):
                pass

            self.handle_events (self.conn.handle_event (event))
            with self._plock:
                event = self.quic.next_event ()

    def handle_events (self, events):
        for event in events:
            if isinstance(event, HeadersReceived):
                self.handle_request (event.stream_id, event.headers, has_data_frame = not event.stream_ended)
                r = self.get_request (event.stream_id)
                if event.stream_ended:
                    r.set_stream_ended ()

            elif isinstance(event, DataReceived) and event.data:
                r = self.get_request (event.stream_id)
                if not r:
                    self.go_away (h3.ErrorCode.HTTP_REQUEST_CANCELLED)
                else:
                    try:
                        r.channel.set_data (event.data, len (event.data))
                    except ValueError:
                        self.go_away (ErrorCodes.HTTP_REQUEST_CANCELLED)

                    if event.stream_ended:
                        if r.collector:
                            r.channel.handle_read ()
                            r.channel.found_terminator ()
                        r.set_stream_ended ()
                        if r.response.is_done ():
                            self.remove_request (event.stream_id)

        self.send_data ()

    def push_promise (self, stream_id, request_headers, addtional_request_headers):
        headers = [(k.encode (), v.encode ()) for k, v in request_headers + addtional_request_headers]
        try:
            promise_stream_id = self.conn.send_push_promise (stream_id = stream_id, headers = headers)
        except NoAvailablePushIDError:
            return
        self.handle_events ([HeadersReceived (headers = headers, stream_ended = True, stream_id = promise_stream_id)])


class Handler (http2_handler.Handler):
    keep_alive = 120
    def match (self, request):
        return request.version.startswith ("3.")

    def handle_request (self, request):
        http3 = http3_request_handler (self, request)
        request.channel.die_with (http3, "http3 connection")
        request.channel.set_socket_timeout (self.keep_alive)
        request.channel.current_request = http3
